//  $Id: PlcEmsGateway.cpp,v 1.6 2009/01/22 11:59:45 tk Exp $
///////////////////////////////////////////////////////////////////////////

//// inclusions ///////////////////////////////////////////////////////////
#include "PlcEmsGateway.h"

#include <iostream>
#include <Messages.h>
#include <sstream>
#include <libV8Plc/Functions.h>
#include <ctype.h>

#include <csignal>

// Use SIGPIPE handler (broken pipes)
// If not, using a socket that is broken will result in an exit
// of the program
bool g_bSigPipeOccured = false;
void catchSigPipe(int)
{
    g_bSigPipeOccured = true;
    std::cout << "ERROR: signal SIGPIPE occured!" << std::endl;
}


//// class CPlcEmsGateway  ////////////////////////////////////////////

//// construction and destruction /////////////////////////////////////////

PLC_CONSTRUCTOR(CPlcEmsGateway)

CPlcEmsGateway::CPlcEmsGateway()
:strSettingsFileName   ( "PowerConsumption.xml" )
{
    m_strServerAddress                  = "192.168.10.44"               ;
    m_iServerPort                       = 23                            ;
    m_dAliveTimeout                     = 0.0                           ;
    m_iAliveCounter                     = 0                             ;
    ResetTelegramBuffer()                                               ;
    m_eReceiveState                     = eReceiveState_WaitSyncStart   ;
    m_bPrintConnectionError             = true                          ;
    m_dLastUpdateTimePowerConsumption   = dGetTime()                    ;
    m_dBurnerPowerOld                   = 0.0                           ;
    m_bBurnerActiveOld                  = false                         ;
    m_dNextIoSaveTime                   = dGetTime() + 600.0            ;
    m_iDayOld                           = 0                             ;
}

/*virtual*/ CPlcEmsGateway::~CPlcEmsGateway()
{
    bStoreIoValues();
    bSocketClose();
}

//// member functions /////////////////////////////////////////////////////

/*virtual*/ void CPlcEmsGateway::CreateIos()
{
    CPlcInterface::CreateIos();

    UseIoPrefix( strGetName() + "_" );

    CIoInteger ioState;
    UINT uiIndex = 0;
    CreateIo( &ioState, "iState" );
    CreateIo( &m_io_In_bConnect             , "In_bConnect"         , &ioState, uiIndex++, true );
    CreateIo( &m_io_Out_bConnected          , "Out_bConnected"      , &ioState, uiIndex++ );
    CreateIo( &m_io_Out_bTelegramError      , "Out_bTelegramError"  , &ioState, uiIndex++ );
    CreateIo( &m_io_In_bPrintDebug          , "In_bPrintDebug"      , &ioState, uiIndex++ );
    CreateIo( &m_io_In_bResetStatistics     , "In_bResetStatistics" , &ioState, uiIndex++ );

    CreateIo( &m_io_Out_strData             , "Out_strData"         );
    CreateIo( &m_io_Out_strTelegram         , "Out_strTelegram"     );
    CreateIo( &m_io_In_strLogFilter         , "In_strLogFilter"     );

    CreateIo( &m_io_In_Server_strAddress    , "In_Server_strAddress", m_strServerAddress);
    CreateIo( &m_io_In_Server_iPort         , "In_Server_iPort"     , m_iServerPort     );
    CreateIo( &m_io_Out_iReceiveState       , "Out_iReceiveState"   );

    CreateIo( &m_io_Out_dDebug1             , "Out_dDebug_1" );
    CreateIo( &m_io_Out_iCrcErrorCounter    , "Out_iCrcErrorCounter" );


    UseIoPrefix( strGetName() + "_Data_RC35_" );
    CreateIo( &m_io_Data_RC35_Out_strDate                       , "Out_strDate"                     );
    CreateIo( &m_io_Data_RC35_Out_strTime                       , "Out_strTime"                     );
    CreateIo( &m_io_Data_RC35_Out_dRoomTemperatureActual        , "Out_dRoomTemperatureActual"      );
    CreateIo( &m_io_Data_RC35_Out_dRoomTemperatureSetpoint      , "Out_dRoomTemperatureSetpoint"    );
    CreateIo( &m_io_Data_RC35_Out_dAmbientTemperatureFiltered   , "Out_dAmbientTemperatureFiltered" );

    CreateIo( &m_io_Data_RC35_Out_iState        , "iOut_iState"         );
    uiIndex = 0;
    CreateIo( &m_io_Data_RC35_Out_Mode_bNight   , "Out_Mode_bNight"     , &m_io_Data_RC35_Out_iState, uiIndex++ );
    CreateIo( &m_io_Data_RC35_Out_Mode_bDay     , "Out_Mode_bDay"       , &m_io_Data_RC35_Out_iState, uiIndex++ );
    CreateIo( &m_io_Data_RC35_Out_Mode_bAuto    , "Out_Mode_bAuto"      , &m_io_Data_RC35_Out_iState, uiIndex++ );

    UseIoPrefix( strGetName() + "_Data_UBA3_" );
    CreateIo( &m_io_Data_UBA3_Out_dFlowTemperatureNom   , "Out_dFlowTemperatureNom"     );
    CreateIo( &m_io_Data_UBA3_Out_dFlowTemperature      , "Out_dFlowTemperature"        );
    CreateIo( &m_io_Data_UBA3_Out_dWaterTemp1           , "Out_dWaterTemp1"             );
    CreateIo( &m_io_Data_UBA3_Out_dReturnTemperature    , "Out_dReturnTemperature"      );
    CreateIo( &m_io_Data_UBA3_Out_dFlameCurrent         , "Out_dFlameCurrent"           );
    CreateIo( &m_io_Data_UBA3_Out_dPressure             , "Out_dPressure"               );
    CreateIo( &m_io_Data_UBA3_Out_AmbientTemperature    , "Out_AmbientTemperature"      );
    CreateIo( &m_io_Data_UBA3_Out_dTemp_00_19_B0607     , "Out_dTemp_00_19_B0607"       );
    CreateIo( &m_io_Data_UBA3_Out_iBurnerStarts         , "Out_iBurnerStarts"           );
    CreateIo( &m_io_Data_UBA3_Out_iBurnerPowerMax       , "Out_iBurnerPowerMax"         );
    CreateIo( &m_io_Data_UBA3_Out_iBurnerPower          , "Out_iBurnerPower"            );
    CreateIo( &m_io_Data_UBA3_Out_dWaterTemp2           , "Out_dWaterTemp2"             );
    CreateIo( &m_io_Data_UBA3_Out_dWaterTemp3           , "Out_dWaterTemp3"             );
    CreateIo( &m_io_Data_UBA3_Out_dWorkingHours         , "Out_dWorkingHours"           );
    CreateIo( &m_io_Data_UBA3_Out_dBurnerWorkingHours   , "Out_dBurnerWorkingHours"     );
    CreateIo( &m_io_Data_UBA3_Out_iErrorCode            , "Out_iErrorCode"              );
    CreateIo( &m_io_Data_UBA3_Out_dWaterTempNom         , "Out_dWaterTempNom"           );
    CreateIo( &m_io_Data_UBA3_Out_dWaterWorkingHours    , "Out_dWaterWorkingHours"      );
    CreateIo( &m_io_Data_UBA3_Out_iMode                 , "Out_iMode"                   );
    CreateIo( &m_io_Data_UBA3_Out_iPumpModulation       , "Out_iPumpModulation"         );


    CreateIo( &m_io_Date_UBA3_Out_iState                , "iOut_iState"                 );
    uiIndex = 0;
    CreateIo( &m_io_Data_UBA3_Out_bCirc                 , "Out_bCirc"                   , &m_io_Date_UBA3_Out_iState, uiIndex++ );
    CreateIo( &m_io_Data_UBA3_Out_bHkWW                 , "Out_bHkWW"                   , &m_io_Date_UBA3_Out_iState, uiIndex++ );
    CreateIo( &m_io_Data_UBA3_Out_bHkPumpe              , "Out_bHkPumpe"                , &m_io_Date_UBA3_Out_iState, uiIndex++ );
    CreateIo( &m_io_Data_UBA3_Out_bBrenner              , "Out_bBrenner"                , &m_io_Date_UBA3_Out_iState, uiIndex++ );
    CreateIo( &m_io_Data_UBA3_Out_bFlameOn              , "Out_bFlameOn"                , &m_io_Date_UBA3_Out_iState, uiIndex++ );
    CreateIo( &m_io_Data_UBA3_Out_bIgnition             , "Out_bIgnition"               , &m_io_Date_UBA3_Out_iState, uiIndex++ );
    CreateIo( &m_io_Data_UBA3_Out_b3wHk1                , "Out_b3wHk1"                  , &m_io_Date_UBA3_Out_iState, uiIndex++ );
    CreateIo( &m_io_Data_UBA3_Out_b3wWW                 , "Out_b3wWW"                   , &m_io_Date_UBA3_Out_iState, uiIndex++ );
    CreateIo( &m_io_Data_UBA3_Out_bModeDay              , "Out_bModeDay"                , &m_io_Date_UBA3_Out_iState, uiIndex++ );
    CreateIo( &m_io_Data_UBA3_Out_bModeHeating          , "Out_bModeHeating"            , &m_io_Date_UBA3_Out_iState, uiIndex++ );
    CreateIo( &m_io_Data_UBA3_Out_bModeNormalOperation  , "Out_bModeNormalOperation"    , &m_io_Date_UBA3_Out_iState, uiIndex++ );

    UseIoPrefix( strGetName() + "_Data_WM10_" );
    CreateIo( &m_io_Data_WM10_Out_dFlowTemperature1     , "Out_dFlowTemperature1"       );
    CreateIo( &m_io_Data_WM10_Out_dFlowTemperature2     , "Out_dFlowTemperature2"       );

    UseIoPrefix( strGetName() + "_Data_" );

    m_oPersistent_Data_In_PowerConsumptionFactor_Water          .Init( this, &m_io_Data_In_PowerConsumptionFactor_Water         , "In_PowerConsumptionFactor_Water"         );
    m_oPersistent_Data_In_PowerConsumptionFactor_Heating        .Init( this, &m_io_Data_In_PowerConsumptionFactor_Heating       , "In_PowerConsumptionFactor_Heating"       );

    m_oPersistent_Data_Out_PowerConsumption_All_Today           .Init( this, &m_io_Data_Out_PowerConsumption_All_Today          , "Out_PowerConsumption_All_Today"          );
    m_oPersistent_Data_Out_PowerConsumption_All_Yesterday       .Init( this, &m_io_Data_Out_PowerConsumption_All_Yesterday      , "Out_PowerConsumption_All_Yesterday"      );
    m_oPersistent_Data_Out_PowerConsumption_All                 .Init( this, &m_io_Data_Out_PowerConsumption_All                , "Out_PowerConsumption_All"                );
    m_oPersistent_Data_Out_PowerConsumption_Water_Today         .Init( this, &m_io_Data_Out_PowerConsumption_Water_Today        , "Out_PowerConsumption_Water_Today"        );
    m_oPersistent_Data_Out_PowerConsumption_Water_Yesterday     .Init( this, &m_io_Data_Out_PowerConsumption_Water_Yesterday    , "Out_PowerConsumption_Water_Yesterday"    );
    m_oPersistent_Data_Out_PowerConsumption_Water_All           .Init( this, &m_io_Data_Out_PowerConsumption_Water_All          , "Out_PowerConsumption_Water"              );
    m_oPersistent_Data_Out_PowerConsumption_Heating_Today       .Init( this, &m_io_Data_Out_PowerConsumption_Heating_Today      , "Out_PowerConsumption_Heating_Today"      );
    m_oPersistent_Data_Out_PowerConsumption_Heating_Yesterday   .Init( this, &m_io_Data_Out_PowerConsumption_Heating_Yesterday  , "Out_PowerConsumption_Heating_Yesterday"  );
    m_oPersistent_Data_Out_PowerConsumption_Heating_All         .Init( this, &m_io_Data_Out_PowerConsumption_Heating_All        , "Out_PowerConsumption_Heating"            );
    m_oPersistent_Data_Out_PowerConsumption_strResetDate        .Init( this, &m_io_Data_Out_PowerConsumption_strResetDate       , "Out_PowerConsumption_strResetDate"       );

}

/*virtual*/ void CPlcEmsGateway::Init()
{
    CPlcInterface::Init();

    // Initialize your I/Os, variables, classes, and so on here.

    // PLC stops if m_bIoInitSuccess == false.

    UseIoPrefix( "" );
    InitIo( &m_io_Setup_strSettingsPath , "Setup_strSettingsPath"   );
    InitIo( &m_io_dCycleStart           , "_dCycleStart"            );
    InitIo( &m_io_Time_strCurrentDate   , "Time_strCurrentDate"     );
    InitIo( &m_io_Time_strCurrentTime   , "Time_strCurrentTime"     );
    InitIo( &m_io_Time_iSeconds         , "Time_iSeconds"           );
    InitIo( &m_io_Time_iDay             , "Time_iDay"               );

    if( bGetIoInitSuccess() )
    {
        m_io_Out_bConnected = false;
        bLoadIoValues();
        m_iDayOld = m_io_Time_iDay;
    }
}

/*virtual*/ void CPlcEmsGateway::Main()
{
    WatchdogFeed( 600.0 );

    CheckDailyStatisticReset();
    CheckResetStatistics();

    if( dGetTime() > m_dNextIoSaveTime  )
    {
        bStoreIoValues();
        m_dNextIoSaveTime = dGetTime() + 600.0; // 10min
    }


    if(g_bSigPipeOccured)
    {
        bSocketClose( );
    }

    if( m_io_In_bConnect )
    {
        if( !m_io_Out_bConnected )
        {
            if( bSocketTimeout() )
            {
                bSocketOpen();
            }
        }
        else
        {
            bSocketReceive();
        }
    }
    else
    {
        if( m_io_Out_bConnected )
        {
            bSocketClose();
        }
    }

    if( !m_io_Out_bConnected )
    {
        ResetPowerConsumptionCalculation();
    }
}

/*virtual*/ void CPlcEmsGateway::SetParameter(
    const CString& strName,
    const CString& strValue)
{
    if( NLibV8Plc::bSetParameter( "strServerAddress", strName, strValue, &m_strServerAddress) ){ return; }
    if( NLibV8Plc::bSetParameter( "iServerPort"     , strName, strValue, &m_iServerPort     ) ){ return; }

    CPlcInterface::SetParameter(strName, strValue);
}

/*virtual*/ const char* CPlcEmsGateway::pszGetMessageText(
    UINT uiMessageId) const
{
//    switch(NLibV8Plc::CMessages::uiGetTypeId(uiMessageId))
//    {
//        case M_INIT_SUCCESS: return "KPlcMessage: Initialization finished; Success=%b0.";
//    }

    return CPlcInterface::pszGetMessageText(uiMessageId);
}

/*virtual*/ bool CPlcEmsGateway::bSocketClose( )
{
    m_io_Out_bConnected = false;
    SocketRestartTimeout(5.0);

    try
    {
        ResetTelegramBuffer();
        m_eReceiveState = eReceiveState_WaitSyncStart;

        m_oSocket.Close();
    }
    catch (std::exception& e)
    {
        PostMessage( CMessages::MS_ERROR, e.what() );
        std::cerr << e.what() << std::endl;
        return false;
    }

    return true;
}

/*virtual*/ bool CPlcEmsGateway::bSocketOpen()
{
    m_io_Out_bConnected = false;
    SocketRestartTimeout();

    try
    {

        if ( ! m_oSocket.create() )
        {
            throw std::logic_error("Could not create client socket." );
        }

        if ( ! m_oSocket.connect ( m_io_In_Server_strAddress.szGetValue(), m_io_In_Server_iPort ) )
        {
            throw std::logic_error ( "Could not bind to port." );
        }
        m_oSocket.set_non_blocking( true);
        m_io_Out_bConnected     = true;
        m_bPrintConnectionError = true;
    }
    catch (std::exception& e)
    {
        if( m_bPrintConnectionError )
        {
            PostMessage( CMessages::MS_ERROR, e.what() );
            m_bPrintConnectionError = false;
        }
        std::cerr << e.what() << std::endl;
        bSocketClose();
    }

    return m_io_Out_bConnected;

}

/*virtual*/ bool CPlcEmsGateway::bSocketSend( CString* pstrBuffer  )
{
    return false;
}

/*virtual*/ bool CPlcEmsGateway::bSocketReceive()
{
    if( !m_io_Out_bConnected )
    {
        return false;
    }

    try
    {
        std::string strBuffer;
        strBuffer.clear();

        int iStatus = m_oSocket.recv( strBuffer );
        m_io_Out_dDebug1 = iStatus;

        if( -1 == iStatus ) // would block
        {
            if( bSocketTimeout() )
            {
                bSocketClose();
            }
            return false;
        }

        if( 0 == iStatus )
        {
            bSocketClose();
            return false; // Connection closed cleanly by peer???
        }

        if( !strBuffer.empty() )
        {
            m_io_Out_strData.SetValue( "" );
            //std::cout << std::endl;

            SocketRestartTimeout();
            for(UINT i=0; i < strBuffer.length(); i++)
            {
                unsigned char chByte = strBuffer.at(i);

                unsigned char cDebugOutByte = ( chByte        == eEOT1 ? ' ' : chByte        );
                              cDebugOutByte = ( cDebugOutByte == eEOT2 ? ' ' : cDebugOutByte );
                m_io_Out_strData.SetValue( CString( m_io_Out_strData.szGetValue () )
                                         + CString( cDebugOutByte ) );
//                 std::cout << chByte; std::cout.flush();

                m_io_Out_iReceiveState = m_eReceiveState;
                switch( m_eReceiveState )
                {
                    case eReceiveState_WaitSyncStart:
                        if( eEOT2 == chByte )
                        {
                            m_eReceiveState = eReceiveState_WaitSync_EOT1;
                        }
                    break;

                    case eReceiveState_WaitSync_EOT1:
                        if( eEOT1 == chByte )
                        {
                            m_eReceiveState = eReceiveState_WaitSync_EOT2;
                        }
                        else
                        {
                            // put data in buffer
                            m_TelegramBuffer[m_ui_TelegramBufferIndex] = chByte;
                            m_ui_TelegramBufferIndex++;
                        }
                    break;

                    case eReceiveState_WaitSync_EOT2:
                        if( eEOT2 == chByte )
                        {
                            // end of telegram -> check length
                            int iTelegramCounter = static_cast<int>( floor( (m_ui_TelegramBufferIndex-2) * 0.5 ) );

                            // check sended data bytes
                            unsigned char szHexDataCount[] = {  m_TelegramBuffer[m_ui_TelegramBufferIndex-2],
                                                                m_TelegramBuffer[m_ui_TelegramBufferIndex-1],
                                                                0 };
                            int iDataCount = iHexToByte( szHexDataCount );

                            if( iTelegramCounter == iDataCount )
                            {
                                m_io_Out_bTelegramError = false;

                                // use only data telegrams
                                TranslateEmsMessage();
                            }
                            else
                            {
                                // std::cout << "length=" << m_ui_TelegramBufferIndex << std::endl;

                                if( m_ui_TelegramBufferIndex > 0 )
                                {
                                    m_io_Out_bTelegramError = true;
                                }
                            }
                            ResetTelegramBuffer();
                            m_eReceiveState = eReceiveState_WaitSync_EOT1;
                        }
                        else
                        {
                            // put prev telegram in buffer
                            m_TelegramBuffer[m_ui_TelegramBufferIndex] = eEOT1;
                            m_ui_TelegramBufferIndex++;
                            if( eEOT1 != chByte )
                            {
                                m_TelegramBuffer[m_ui_TelegramBufferIndex] = chByte;
                                m_ui_TelegramBufferIndex++;
                                m_eReceiveState = eReceiveState_WaitSync_EOT1;
                            }
                        }
                    break;
                }
            }
            return true;
        }
    }
    catch (std::exception& e)
    {
        PostMessage( CMessages::MS_ERROR, e.what() );
        std::cerr << e.what() << std::endl;
        bSocketClose();
        return false;
    }

    return false;
}

/*virtual*/ void CPlcEmsGateway::SocketRestartTimeout( double dTimeout )
{
    m_dAliveTimeout = dGetTime() + dTimeout;
}

/*virtual*/ int CPlcEmsGateway::iHexToByte(unsigned char *szHex)
{
    int x = 0;
    int y = 1;

    for( int i=1; i>=0; i-- )
    {
        if('0' <= szHex[i] && szHex[i]<= '9')
        {
            x+=y*(szHex[i] - '0');
        }
        else if('a' <= szHex[i] && szHex[i] <= 'f')
        {
            x+=y*(szHex[i] - 'a' + 10);
        }
        else if('A' <= szHex[i] && szHex[i] <= 'F')
        {
            x+=y*(10 + szHex[i] - 'A');
        }
        else
        {
            return -1; /* invalid input! */
        }
        y<<=4;
    }
    return x;
}

/*virtual*/ bool CPlcEmsGateway::bHexToByteBuffer(unsigned char *szHex, unsigned char *szByte)
{
    ASSERT( szHex );
    ASSERT( szByte);
    int iCharBufferPointer = 0;
    szByte[0] = '\0';

    if( 0 == strlen((char*)szHex) || eMaxBufferSize < strlen((char*)szHex) )
    {
        return false;
    }

    for(UINT i=0; i<strlen((char*)szHex); i+=2)
    {
        unsigned char szHexByteData[]={ szHex[i], szHex[i+1], 0};
        int iData = iHexToByte( szHexByteData  );
        if( iData >= 0 )
        {
            szByte[iCharBufferPointer] = static_cast<unsigned char>( iData );
            iCharBufferPointer++;
        }
    }
    szByte[iCharBufferPointer] = 0;
    return true;
}

/*virtual*/ void CPlcEmsGateway::TranslateEmsMessage()
{
    // DEBUG
    //unsigned char szTestMsg[] ="0800180007027600000000006280000215027E00000D304800CB000000191E"; // OK CRC
    //memcpy(m_TelegramBuffer, szTestMsg, sizeof(szTestMsg));

    unsigned char szDataBuffer[eMaxBufferSize+1];
    memset(szDataBuffer, 0, sizeof(szDataBuffer) );

    int iTelegramLength = strlen( (char*)m_TelegramBuffer ) / 2;
    if( bHexToByteBuffer(m_TelegramBuffer, szDataBuffer) )
    {
        m_io_Out_strTelegram.SetValue( (char*)m_TelegramBuffer );
        if( m_io_In_bPrintDebug )
        {
            PrintTelegramBuffer( m_TelegramBuffer );
        }

        if(iTelegramLength > 2)
        {
            const unsigned char ucCrcReceived = szDataBuffer[iTelegramLength-2];
            szDataBuffer[iTelegramLength-2] = 0; //remove crc
            const unsigned char ucCrcCalc = ucCrc(szDataBuffer, iTelegramLength-2);
            if( ucCrcCalc == ucCrcReceived)
            {
                UpdateIoInterface( szDataBuffer );
            }
            else
            {
                m_io_Out_iCrcErrorCounter = m_io_Out_iCrcErrorCounter + 1;
            }
        }
    }
}

/*virtual*/ void CPlcEmsGateway::PrintTelegramBuffer( unsigned char* szBuffer )
{
    // format telegram with underscores
    CString strPrintTelegramm( szBuffer );
    CString strFilter = m_io_In_strLogFilter.szGetValue();

    bool bPrint = strFilter.bIsEmpty()
               || ( 0 == strFilter.iCompareNoCase( strPrintTelegramm.strLeft( strFilter.uiGetLength() ) ) );

    if(!bPrint)
    {
        return;
    }

    strPrintTelegramm.Empty();
    int iCharCount=0;

    for(UINT i=0; i<strlen( (char*) szBuffer ); i++)
    {

        strPrintTelegramm = strPrintTelegramm + szBuffer[i];
        iCharCount++;
        if(iCharCount >= 2)
        {
            strPrintTelegramm = strPrintTelegramm + "_";
            iCharCount = 0;
        }
    }

    double dTime_10ms = fmod( dGetTime(), 1.0 ) * 100.0;


    std::cout   << m_io_Time_strCurrentDate.szGetValue() << " "
                << m_io_Time_strCurrentTime.szGetValue() << "."
                << rint( dTime_10ms )                    << " # "
                << strPrintTelegramm.szGetBuffer()       << std::endl;
}


// *virtual* void CPlcEmsGateway::ConvertToHex( const  unsigned char* szBuffer, CString* pstrHexMsg, bool bPrint )
// {
//     ASSERT( pstrHexMsg );
//
//     *pstrHexMsg = "";
//     for(UINT i=0; i<strlen( reinterpret_cast<const unsigned char*>(szBuffer) ); i++)
//     {
//         // strHexText = strHexText + CString( unsigned char(m_TelegramBuffer[i]) );
//
//         CString strHexValue = strIntToHex(int(m_TelegramBuffer[i]));
//         strHexValue = strHexValue.strRight( 3 ); // unsigned char value and space
//         *pstrHexMsg = *pstrHexMsg + strHexValue;
//     }
//
//     if( bPrint )
//     {
//         //std::cout << m_TelegramBuffer << std::endl;
//         std::cout << pstrHexMsg->szGetBuffer() << std::endl;
//     }
// }

/*virtual*/ void CPlcEmsGateway::ResetTelegramBuffer()
{
    m_ui_TelegramBufferIndex = 0;
    memset( m_TelegramBuffer, 0, sizeof( m_TelegramBuffer ) );
}

/*virtual*/ void CPlcEmsGateway::UpdateIoInterface( unsigned char* szMsg )
{
    if( eRC35 == szMsg[0] )
    {
        Decode_RC35( szMsg );
    }
    else if( eUBA3 == szMsg[0] )
    {
        Decode_UBA3( szMsg );
    }
    else if( eWM10 == szMsg[0] )
    {
        Decode_WM10( szMsg );
    }


}

/*virtual*/ void CPlcEmsGateway::Decode_RC35( unsigned char* szMsg )
{
    if      ( 0x0 == szMsg[1] && 0x6 == szMsg[2] )
    {
        Decode_RC35_00_06_Systemdate( szMsg );
    }
    else if ( 0x0 == szMsg[1] && 0xA3 == szMsg[2] )
    {
        Decode_RC35_00_A3_Temperature( szMsg );
    }
    else if ( 0x0 == szMsg[1] && 0x3D == szMsg[2] )
    {
        Decode_RC35_00_3D_Setpoints( szMsg );
    }
    else if ( 0x0 == szMsg[1] && 0x3E == szMsg[2] )
    {
        Decode_RC35_00_3E_Roomtemperature( szMsg );
    }
}

/*virtual*/ void CPlcEmsGateway::Decode_UBA3( unsigned char* szMsg )
{
    if( 0x18 == szMsg[2] )
    {
        Decode_UBA3_00_18_Temperatures( szMsg );
    }
    else if( 0x19 == szMsg[2] )
    {
        Decode_UBA3_00_19_TempOutside_Counters( szMsg );
    }
    else if( 0x34 == szMsg[2] )
    {
        Decode_UBA3_00_34_Water( szMsg );
    }
    else if( 0x10 == szMsg[1] && 0x14 == szMsg[2] )
    {
        Decode_UBA3_10_14_WorkingHours( szMsg );
    }
}

/*virtual*/ void CPlcEmsGateway::Decode_WM10( unsigned char* szMsg )
{
    if( 0x9C == szMsg[2] )
    {
        Decode_WM10_11_00_9C_Temperature( szMsg );
    }
    else if( 0x1E == szMsg[2] )
    {
        Decode_WM10_11_08_1E_Temperature( szMsg );
    }
}




/*virtual*/ void CPlcEmsGateway::Decode_RC35_00_06_Systemdate( unsigned char* szMsg )
{

    unsigned char szDay[][3] ={ "Mo", "Di", "Mi", "Do", "Fr", "Sa", "So" };
    int iDayIndex = tMinMax<int>( szMsg[10], 0, 6);

    //CString strDate = CString(&szDay[iDayIndex][0]) + CString( " %d-%02d-%02d", szMsg[4]+2000, szMsg[5], szMsg[7] );
    CString strDate( "%s %d-%02d-%02d"  , &szDay[iDayIndex][0], szMsg[4]+2000, szMsg[5], szMsg[7] );
    CString strTime( "%02d:%02d:%02d"   , szMsg[6], szMsg[8], szMsg[9] );

    m_io_Data_RC35_Out_strDate            .SetValue ( strDate );
    m_io_Data_RC35_Out_strTime            .SetValue ( strTime );
}

/*virtual*/ void CPlcEmsGateway::Decode_RC35_00_3D_Setpoints( unsigned char* szMsg )
{
    if(  0x02 == szMsg[3] )
    {
        // double dRaumTempGest = static_cast<double>( szMsg[4] ) / 2.0;
    }
    else if( 0x07 == szMsg[3] )
    {
        m_io_Data_RC35_Out_Mode_bNight  = szMsg[4] && (1 << 0);
        m_io_Data_RC35_Out_Mode_bDay    = szMsg[4] && (1 << 1);
        m_io_Data_RC35_Out_Mode_bAuto   = szMsg[4] && (1 << 2);
    }
}

/*virtual*/ void CPlcEmsGateway::Decode_RC35_00_A3_Temperature(  unsigned char* szMsg )
{
    m_io_Data_RC35_Out_dAmbientTemperatureFiltered = tConvertShortValue<double>(szMsg[4], szMsg[5]);
}


/*virtual*/ void CPlcEmsGateway::Decode_RC35_00_3E_Roomtemperature(  unsigned char* szMsg )
{
    m_io_Data_RC35_Out_dRoomTemperatureSetpoint = static_cast<double>( szMsg[6] ) / 2.0;
    m_io_Data_RC35_Out_dRoomTemperatureActual   = tConvertShortValue<double>(szMsg[7], szMsg[8]) / 10.0;
}



/*virtual*/ void CPlcEmsGateway::Decode_UBA3_00_18_Temperatures( unsigned char* szMsg )
{
    m_io_Data_UBA3_Out_dFlowTemperatureNom  = tConvertByteValue <double>(szMsg[ 4]);
    m_io_Data_UBA3_Out_dFlowTemperature     = tConvertShortValue<double>(szMsg[ 5], szMsg[ 6]) / 10.0 ;
    m_io_Data_UBA3_Out_iBurnerPowerMax      = tConvertByteValue <int>   (szMsg[ 7]);
    m_io_Data_UBA3_Out_iBurnerPower         = tConvertByteValue <int>   (szMsg[ 8]);
    m_io_Data_UBA3_Out_dWaterTemp1          = tConvertShortValue<double>(szMsg[15], szMsg[16]) / 10.0 ;
    m_io_Data_UBA3_Out_dReturnTemperature   = tConvertShortValue<double>(szMsg[17], szMsg[18]) / 10.0 ;
    m_io_Data_UBA3_Out_dFlameCurrent        = tConvertShortValue<double>(szMsg[19], szMsg[20]) / 10.0 ;
    m_io_Data_UBA3_Out_dPressure            = tConvertByteValue <double>(szMsg[21]           ) / 10.0 ;
    m_io_Data_UBA3_Out_iErrorCode           = tConvertByteValue <int>   (szMsg[25]);

    unsigned char chStatus = szMsg[11];
    m_io_Data_UBA3_Out_bFlameOn     = chStatus & ( 1 <<  0 );
    m_io_Data_UBA3_Out_bBrenner     = chStatus & ( 1 <<  2 );
    m_io_Data_UBA3_Out_bIgnition    = chStatus & ( 1 <<  3 );
    m_io_Data_UBA3_Out_bHkPumpe     = chStatus & ( 1 <<  5 );
    m_io_Data_UBA3_Out_bHkWW        = chStatus & ( 1 <<  6 );
    m_io_Data_UBA3_Out_bCirc        = chStatus & ( 1 <<  7 );
    m_io_Data_UBA3_Out_b3wHk1       = chStatus & ( 1 <<  8 );
    m_io_Data_UBA3_Out_b3wWW        = chStatus & ( 1 << 12 );

    UpdatePowerConsumption();
}

/*virtual*/ void CPlcEmsGateway::Decode_UBA3_00_19_TempOutside_Counters( unsigned char* szMsg )
{
    m_io_Data_UBA3_Out_AmbientTemperature   = tConvertShortValue<double>(szMsg[ 4], szMsg[ 5]) / 10.0;
    m_io_Data_UBA3_Out_dTemp_00_19_B0607    = tConvertShortValue<double>(szMsg[ 6], szMsg[ 7]) / 10.0;
    m_io_Data_UBA3_Out_iPumpModulation      = tConvertByteValue <int>   (szMsg[13]);
    m_io_Data_UBA3_Out_iBurnerStarts        = tConvertIntValue<int>  (0, szMsg[14], szMsg[15], szMsg[16]);
    m_io_Data_UBA3_Out_dBurnerWorkingHours  = tConvertIntValue<int>  (0, szMsg[17], szMsg[18], szMsg[19]) / 60.0;
}

/*virtual*/ void CPlcEmsGateway::Decode_UBA3_00_34_Water( unsigned char* szMsg )
{
    m_io_Data_UBA3_Out_dWaterTempNom        = tConvertByteValue <double>(szMsg[ 4] );
    m_io_Data_UBA3_Out_dWaterTemp2          = tConvertShortValue<double>(szMsg[ 5], szMsg[ 6]) / 10.0;
    m_io_Data_UBA3_Out_dWaterTemp3          = tConvertShortValue<double>(szMsg[ 7], szMsg[ 8]) / 10.0;
    m_io_Data_UBA3_Out_dWaterWorkingHours   = tConvertIntValue<int>  (0, szMsg[14], szMsg[15], szMsg[16]) / 60.0;

    unsigned char chStatus = szMsg[9];
    m_io_Data_UBA3_Out_iMode                = chStatus;
    m_io_Data_UBA3_Out_bModeDay             = chStatus & ( 1 << 0);
    m_io_Data_UBA3_Out_bModeHeating         = chStatus & ( 1 << 3);
    m_io_Data_UBA3_Out_bModeNormalOperation = chStatus & ( 1 << 5);

}

/*virtual*/ void CPlcEmsGateway::Decode_UBA3_10_14_WorkingHours( unsigned char* szMsg )
{
    // Die Minuten muessen dann noch nach 60 umgerechnet werden ...
    m_io_Data_UBA3_Out_dWorkingHours = tConvertIntValue<double>(0, szMsg[4], szMsg[5], szMsg[6]) / 60.0;
}


/*virtual*/ void CPlcEmsGateway::Decode_WM10_11_00_9C_Temperature( unsigned char* szMsg )
{
    m_io_Data_WM10_Out_dFlowTemperature1 = tConvertShortValue<double>(szMsg[ 4], szMsg[ 5]) / 10.0;
}

/*virtual*/ void CPlcEmsGateway::Decode_WM10_11_08_1E_Temperature( unsigned char* szMsg )
{
    m_io_Data_WM10_Out_dFlowTemperature2 = tConvertShortValue<double>(szMsg[ 4], szMsg[ 5]) / 10.0;
}

/*virtual*/ unsigned char CPlcEmsGateway::ucCrc( unsigned char* szTelegram, int iLength)
{
    unsigned char crc1 = 0;

    for(int i=0; i<iLength;i++)
    {
        int d=0;
        if( 0x0 != (crc1 & 0x80))
        {
            crc1 = (crc1^12);
            d=1;
        }
        crc1 = (crc1 << 1);
        crc1 &= 0xfe;
        crc1 |= d;
        crc1 = (crc1 ^ szTelegram[i]);
    }
    return crc1;
}

/*virtual*/ void CPlcEmsGateway::UpdatePowerConsumption()
{
    // use this internal values for calculation
    double dTimeDelta       = dGetTime() - m_dLastUpdateTimePowerConsumption;
    double dBurnerPowerOld  = m_dBurnerPowerOld;
    bool   bBurnerActiveOld = m_bBurnerActiveOld;

    // store actual state
    m_dLastUpdateTimePowerConsumption   = dGetTime()                        ;
    m_bBurnerActiveOld                  = m_io_Data_UBA3_Out_bBrenner       ;
    m_dBurnerPowerOld                   = m_io_Data_UBA3_Out_iBurnerPower   ;

    if( !bBurnerActiveOld )
    {
        // burner was off
        return;
    }
    bool bWaterActive      = m_io_Data_UBA3_Out_bHkWW;
    const double dMaxPower = 15.0; // GB162 15kW
    const double dAvgPower = 0.5 * ( dBurnerPowerOld + m_io_Data_UBA3_Out_iBurnerPower );

    double dPowerConsumption = dMaxPower * dAvgPower * 0.01 ; // 100% - 0..1
    dPowerConsumption       *= ( dTimeDelta / 3600.0 )      ; // burner time in [s] -> [h]
    dPowerConsumption       *= ( bWaterActive ?  m_io_Data_In_PowerConsumptionFactor_Water : m_io_Data_In_PowerConsumptionFactor_Heating );


    m_io_Data_Out_PowerConsumption_All               += dPowerConsumption;
    m_io_Data_Out_PowerConsumption_All_Today         += dPowerConsumption;

    if( bWaterActive )
    {
        m_io_Data_Out_PowerConsumption_Water_Today   += dPowerConsumption;
        m_io_Data_Out_PowerConsumption_Water_All     += dPowerConsumption;
    }
    else
    {
        m_io_Data_Out_PowerConsumption_Heating_Today += dPowerConsumption;
        m_io_Data_Out_PowerConsumption_Heating_All   += dPowerConsumption;
    }
}

/*virtual*/ void CPlcEmsGateway::ResetPowerConsumptionCalculation()
{
    m_dLastUpdateTimePowerConsumption   = dGetTime();
    m_dBurnerPowerOld                   = 0.0;
    m_bBurnerActiveOld                  = false;
}

/*virtual*/ bool CPlcEmsGateway::bStoreIoValues()
{
    if(    m_oPersistent_Data_Out_PowerConsumption_All_Today        .bNeedsUpdate()
        || m_oPersistent_Data_Out_PowerConsumption_All_Yesterday    .bNeedsUpdate()
        || m_oPersistent_Data_Out_PowerConsumption_All              .bNeedsUpdate()
        || m_oPersistent_Data_Out_PowerConsumption_Water_Today      .bNeedsUpdate()
        || m_oPersistent_Data_Out_PowerConsumption_Water_Yesterday  .bNeedsUpdate()
        || m_oPersistent_Data_Out_PowerConsumption_Water_All        .bNeedsUpdate()
        || m_oPersistent_Data_Out_PowerConsumption_Heating_Today    .bNeedsUpdate()
        || m_oPersistent_Data_Out_PowerConsumption_Heating_Yesterday.bNeedsUpdate()
        || m_oPersistent_Data_Out_PowerConsumption_Heating_All      .bNeedsUpdate()
        || m_oPersistent_Data_Out_PowerConsumption_strResetDate     .bNeedsUpdate()
        || m_oPersistent_Data_In_PowerConsumptionFactor_Water       .bNeedsUpdate()
        || m_oPersistent_Data_In_PowerConsumptionFactor_Heating     .bNeedsUpdate()
    )
    {
        CParameter oParameter;
        oParameter.bLoad( m_io_Setup_strSettingsPath.strGetValue() + strSettingsFileName );

        m_oPersistent_Data_Out_PowerConsumption_All_Today           .StoreToFile(&oParameter);
        m_oPersistent_Data_Out_PowerConsumption_All_Yesterday       .StoreToFile(&oParameter);
        m_oPersistent_Data_Out_PowerConsumption_All                 .StoreToFile(&oParameter);
        m_oPersistent_Data_Out_PowerConsumption_Water_Today         .StoreToFile(&oParameter);
        m_oPersistent_Data_Out_PowerConsumption_Water_Yesterday     .StoreToFile(&oParameter);
        m_oPersistent_Data_Out_PowerConsumption_Water_All           .StoreToFile(&oParameter);
        m_oPersistent_Data_Out_PowerConsumption_Heating_Today       .StoreToFile(&oParameter);
        m_oPersistent_Data_Out_PowerConsumption_Heating_Yesterday   .StoreToFile(&oParameter);
        m_oPersistent_Data_Out_PowerConsumption_Heating_All         .StoreToFile(&oParameter);
        m_oPersistent_Data_Out_PowerConsumption_strResetDate        .StoreToFile(&oParameter);
        m_oPersistent_Data_In_PowerConsumptionFactor_Water          .StoreToFile(&oParameter);
        m_oPersistent_Data_In_PowerConsumptionFactor_Heating        .StoreToFile(&oParameter);

        return oParameter.bSave( m_io_Setup_strSettingsPath.strGetValue() + strSettingsFileName );
    }
    return true;
}

/*virtual*/ bool CPlcEmsGateway::bLoadIoValues()
{
    CParameter oParameter;
    bool bSuccess = oParameter.bLoad( m_io_Setup_strSettingsPath.strGetValue() + strSettingsFileName );
    if(bSuccess)
    {
        PostMessage(CMessages::MS_INFO, "Intialize Power consumption counters from file");

        m_oPersistent_Data_Out_PowerConsumption_All_Today           .LoadFromFile(&oParameter, 0.0 );
        m_oPersistent_Data_Out_PowerConsumption_All_Yesterday       .LoadFromFile(&oParameter, 0.0 );
        m_oPersistent_Data_Out_PowerConsumption_All                 .LoadFromFile(&oParameter, 0.0 );
        m_oPersistent_Data_Out_PowerConsumption_Water_Today         .LoadFromFile(&oParameter, 0.0 );
        m_oPersistent_Data_Out_PowerConsumption_Water_Yesterday     .LoadFromFile(&oParameter, 0.0 );
        m_oPersistent_Data_Out_PowerConsumption_Water_All           .LoadFromFile(&oParameter, 0.0 );
        m_oPersistent_Data_Out_PowerConsumption_Heating_Today       .LoadFromFile(&oParameter, 0.0 );
        m_oPersistent_Data_Out_PowerConsumption_Heating_Yesterday   .LoadFromFile(&oParameter, 0.0 );
        m_oPersistent_Data_Out_PowerConsumption_Heating_All         .LoadFromFile(&oParameter, 0.0 );
        m_oPersistent_Data_Out_PowerConsumption_strResetDate        .LoadFromFile(&oParameter, CString("unknown") );
        m_oPersistent_Data_In_PowerConsumptionFactor_Water          .LoadFromFile(&oParameter, 1.02 );
        m_oPersistent_Data_In_PowerConsumptionFactor_Heating        .LoadFromFile(&oParameter, 1.0  );
    }
    return bSuccess;
}

/*virtual*/ void CPlcEmsGateway::CheckDailyStatisticReset()
{
    if( m_io_Time_iDay != 0 && 0 == m_iDayOld )
    {
        m_iDayOld = m_io_Time_iDay;
    }
    if( m_iDayOld != m_io_Time_iDay )
    {
        m_io_Data_Out_PowerConsumption_All_Yesterday        = m_io_Data_Out_PowerConsumption_All_Today      ;
        m_io_Data_Out_PowerConsumption_Water_Yesterday      = m_io_Data_Out_PowerConsumption_Water_Today    ;
        m_io_Data_Out_PowerConsumption_Heating_Yesterday    = m_io_Data_Out_PowerConsumption_Heating_Today  ;

        m_io_Data_Out_PowerConsumption_All_Today            = 0.0;
        m_io_Data_Out_PowerConsumption_Water_Today          = 0.0;
        m_io_Data_Out_PowerConsumption_Heating_Today        = 0.0;

        m_iDayOld = m_io_Time_iDay;
    }
}

/*virtual*/ void CPlcEmsGateway::CheckResetStatistics()
{
    if( !m_io_In_bResetStatistics )
    {
        return;
    }

    m_io_In_bResetStatistics = false;

    CString strDateTime("%s - %s"   , m_io_Time_strCurrentDate.szGetValue()
                                    , m_io_Time_strCurrentTime.szGetValue() );

    m_io_Data_Out_PowerConsumption_All_Yesterday        = 0.0;
    m_io_Data_Out_PowerConsumption_Water_Yesterday      = 0.0;
    m_io_Data_Out_PowerConsumption_Heating_Yesterday    = 0.0;
    m_io_Data_Out_PowerConsumption_All_Today            = 0.0;
    m_io_Data_Out_PowerConsumption_Water_Today          = 0.0;
    m_io_Data_Out_PowerConsumption_Heating_Today        = 0.0;

    m_io_Data_Out_PowerConsumption_strResetDate.SetValue( strDateTime.szGetBuffer() );

    PostMessage(CMessages::MS_WARNING, "Reset Power Consumption" );
    CString strMsgText;

    strMsgText.Format("PowerConsumption all  = %2.1f"       , m_io_Data_Out_PowerConsumption_All.dGetValue() );
    PostMessage(CMessages::MS_INFO, strMsgText);

    strMsgText.Format("PowerConsumption water  = %2.1f"     , m_io_Data_Out_PowerConsumption_Water_All.dGetValue() );
    PostMessage(CMessages::MS_INFO, strMsgText);

    strMsgText.Format("PowerConsumption heating  = %2.1f"  , m_io_Data_Out_PowerConsumption_Heating_All.dGetValue() );
    PostMessage(CMessages::MS_INFO, strMsgText);

    m_io_Data_Out_PowerConsumption_All              = 0.0;
    m_io_Data_Out_PowerConsumption_Water_All        = 0.0;
    m_io_Data_Out_PowerConsumption_Heating_All      = 0.0;

}

// end of file
