//  $Id: PlcEmsGateway.h,v 1.4 2008/11/06 17:40:38 tk Exp $
//
//  Declaration of class CPlcEmsGateway.

///////////////////////////////////////////////////////////////////////////
#if !defined(_PLCEMSGATEWAY_H_)
#define _PLCEMSGATEWAY_H_

//// inclusions ///////////////////////////////////////////////////////////

#include <Aradex/Aradex.h>
#include <libV8Plc/ImplicitNamespaces.h>
#include <libV8User/PlcInterface.h>
#include <libV8User/PlcMain.h>
#include <string>

#include <iostream>
#include "Socket.h"
#include "PersistentIo.h"

//// class CPlcEmsGateway /////////////////////////////////////////////

class CPlcEmsGateway : public CPlcInterface
{
public:

    //// construction and destruction /////////////////////////////////////

    // PURPOSE
    //  Constructor.
    CPlcEmsGateway();

    // PURPOSE
    //  Destructor.
    virtual ~CPlcEmsGateway();

    //// member functions /////////////////////////////////////////////////

    // PURPOSE
    //  Overrides the CPlcInterface method.
    virtual void CreateIos();

    // PURPOSE
    //  Overrides the CPlcInterface method.
    virtual void Init();

    // PURPOSE
    //  Update-method for this PLC. Will be periodically called by
    //  parent process. Implement the functionality of your PLC here.
    virtual void Main();

    // PURPOSE
    //  Is called for every parameter passed to this PLC
    // PARAMETERS
    //  strName:    The parameter name.
    //  strValue:   The parameter value.
    // RETURN VALUE
    //  'true' on success or 'false' otherwise.
    virtual void SetParameter(
        const CString& strName,
        const CString& strValue);

    // PURPOSE
    //  Overrides the CPlcInterface method.
    virtual const char* pszGetMessageText(UINT uiMessageId) const;

    virtual unsigned char ucCrc( unsigned char* szTelegram, int iLength);

    // EMS bus address
    enum { eUBA3 = 0x8  };
    enum { eBC10 = 0x9  };
    enum { eRC35 = 0x10 };
    enum { eWM10 = 0x11 };
    enum { eMM10 = 0x21 };

protected:

    template< typename T >
    inline const T tConvertIntValue(const unsigned char& Byte3, const unsigned char& Byte2, const unsigned char& Byte1, const unsigned char& Byte0 )
    {
        unsigned int iByte3 = ( static_cast<unsigned short>(Byte3) & 0xFF ) << 24;
        unsigned int iByte2 = ( static_cast<unsigned short>(Byte2) & 0xFF ) << 16;
        unsigned int iByte1 = ( static_cast<unsigned short>(Byte1) & 0xFF ) <<  8;
        unsigned int iByte0 = ( static_cast<unsigned short>(Byte0) & 0xFF );

        return ( iByte3 | iByte2 | iByte1 | iByte0 ) & 0xFFFFFFFF;
    }

    template< typename T >
    inline const T tConvertShortValue(const unsigned char& HighByte, const unsigned char& LowByte )
    {
        unsigned short usHighByte = ( static_cast<unsigned short>(HighByte) & 0xFF ) << 8 ;
        unsigned short usLowByte  = ( static_cast<unsigned short>(LowByte ) & 0xFF );
        unsigned short usValue    = ( usHighByte | usLowByte ) & 0xFFFF;

        if( usValue & 0x8000 )
        {
            // convert from two's complement
            usValue &= 0x7FFF;
            usValue -= 1;
            usValue ^= 0x7FFF;
            return usValue * -1;
        }

        return usValue;
    }

    template< typename T >
    inline const T tConvertByteValue(const char& Byte)
    {
        return static_cast<unsigned char>( Byte ) & 0xFF;
    }


    virtual bool bSocketOpen();
    virtual bool bSocketClose();
    virtual bool bSocketSend( CString* pstrBuffer  );
    virtual bool bSocketReceive();
    virtual void SocketRestartTimeout( double dTimeout = 60.0 );
    virtual bool bSocketTimeout() { return m_io_dCycleStart > m_dAliveTimeout;}

    virtual void TranslateEmsMessage();
    virtual void ResetTelegramBuffer();
    virtual int  iHexToByte(unsigned char *szHex);
    virtual bool bHexToByteBuffer(unsigned char *szHex, unsigned char *szByte);
    //virtual void ConvertToHex( const unsigned char* szBuffer, CString* pstrHexMsg, bool bPrint );
    virtual void PrintTelegramBuffer( unsigned char* szBuffer );
    virtual void UpdateIoInterface( unsigned char* szMsg );

    virtual void Decode_RC35( unsigned char* szMsg );
    virtual void Decode_UBA3( unsigned char* szMsg );
    virtual void Decode_WM10( unsigned char* szMsg );

    virtual void Decode_RC35_00_06_Systemdate       ( unsigned char* szMsg );
    virtual void Decode_RC35_00_3D_Setpoints        ( unsigned char* szMsg );
    virtual void Decode_RC35_00_3E_Roomtemperature  ( unsigned char* szMsg );
    virtual void Decode_RC35_00_A3_Temperature      ( unsigned char* szMsg );

    virtual void Decode_UBA3_00_18_Temperatures     ( unsigned char* szMsg );
    virtual void Decode_UBA3_00_19_TempOutside_Counters( unsigned char* szMsg );
    virtual void Decode_UBA3_00_34_Water            ( unsigned char* szMsg );
    virtual void Decode_UBA3_10_14_WorkingHours     ( unsigned char* szMsg );

    virtual void Decode_WM10_11_00_9C_Temperature   ( unsigned char* szMsg );
    virtual void Decode_WM10_11_08_1E_Temperature   ( unsigned char* szMsg );

    virtual void UpdatePowerConsumption();
    virtual void ResetPowerConsumptionCalculation();

    virtual bool bStoreIoValues();
    virtual bool bLoadIoValues();

    virtual void CheckDailyStatisticReset();
    virtual void CheckResetStatistics();

    //// log message identifiers //////////////////////////////////////////

    enum EMessages
    {
    };

    //// data members /////////////////////////////////////////////////////

    enum { eMaxBufferSize = 200 };
    const CString       strSettingsFileName                             ;
    CString             m_strServerAddress                              ;
    int                 m_iServerPort                                   ;

    CIoString           m_io_Setup_strSettingsPath                      ;
    CIoString           m_io_Time_strCurrentDate                        ;
    CIoString           m_io_Time_strCurrentTime                        ;
    CIoInteger          m_io_Time_iSeconds                              ;
    CIoInteger          m_io_Time_iDay                                  ;
    int                 m_iDayOld                                       ;

    unsigned char       m_TelegramBuffer[eMaxBufferSize+1]              ;
    UINT                m_ui_TelegramBufferIndex                        ;
    CIoDouble           m_io_dCycleStart                                ;
    double              m_dAliveTimeout                                 ;
    int                 m_iAliveCounter                                 ;

    CSocket             m_oSocket                                       ;
    bool                m_bPrintConnectionError                         ;

    CIoBoolean          m_io_In_bConnect                                ;
    CIoBoolean          m_io_Out_bConnected                             ;
    CIoString           m_io_Out_strData                                ;
    CIoString           m_io_Out_strTelegram                            ;
    CIoString           m_io_In_Server_strAddress                       ;
    CIoInteger          m_io_In_Server_iPort                            ;
    CIoString           m_io_In_strLogFilter                            ;
    CIoBoolean          m_io_In_bResetStatistics                        ;

    enum EReceiveState
    {
        eReceiveState_WaitSyncStart ,
        eReceiveState_WaitSync_EOT1 ,
        eReceiveState_WaitSync_EOT2
    };

    enum EEndOfTelegram
    {
        eEOT1 = 0xD, // \r
        eEOT2 = 0xA  // \n
    };

    EReceiveState       m_eReceiveState                                 ;
    CIoInteger          m_io_Out_iReceiveState                          ;
    CIoBoolean          m_io_Out_bTelegramError                         ;
    CIoDouble           m_io_Out_dDebug1                                ;
    CIoBoolean          m_io_In_bPrintDebug                             ;
    CIoInteger          m_io_Out_iCrcErrorCounter                       ;

    // EMS Data
    CIoString           m_io_Data_RC35_Out_strDate                      ;
    CIoString           m_io_Data_RC35_Out_strTime                      ;
    CIoDouble           m_io_Data_RC35_Out_dRoomTemperatureActual       ;
    CIoDouble           m_io_Data_RC35_Out_dRoomTemperatureSetpoint     ;
    CIoDouble           m_io_Data_RC35_Out_dAmbientTemperatureFiltered  ;
    CIoInteger          m_io_Data_RC35_Out_iState                       ;
    CIoBoolean          m_io_Data_RC35_Out_Mode_bNight                  ;
    CIoBoolean          m_io_Data_RC35_Out_Mode_bDay                    ;
    CIoBoolean          m_io_Data_RC35_Out_Mode_bAuto                   ;


    CIoDouble           m_io_Data_UBA3_Out_dFlowTemperatureNom          ;
    CIoDouble           m_io_Data_UBA3_Out_dFlowTemperature             ;
    CIoDouble           m_io_Data_UBA3_Out_dWaterTemp1                  ;
    CIoDouble           m_io_Data_UBA3_Out_dReturnTemperature           ;
    CIoDouble           m_io_Data_UBA3_Out_dFlameCurrent                ;
    CIoDouble           m_io_Data_UBA3_Out_dPressure                    ;

    CIoInteger          m_io_Date_UBA3_Out_iState                       ;
    CIoBoolean          m_io_Data_UBA3_Out_bCirc                        ;
    CIoBoolean          m_io_Data_UBA3_Out_bHkWW                        ;
    CIoBoolean          m_io_Data_UBA3_Out_bHkPumpe                     ;
    CIoBoolean          m_io_Data_UBA3_Out_bBrenner                     ;
    CIoBoolean          m_io_Data_UBA3_Out_bFlameOn                     ;
    CIoBoolean          m_io_Data_UBA3_Out_bIgnition                    ;
    CIoBoolean          m_io_Data_UBA3_Out_b3wHk1                       ;
    CIoBoolean          m_io_Data_UBA3_Out_b3wWW                        ;
    CIoDouble           m_io_Data_UBA3_Out_AmbientTemperature           ;
    CIoDouble           m_io_Data_UBA3_Out_dTemp_00_19_B0607            ;
    CIoInteger          m_io_Data_UBA3_Out_iBurnerStarts                ;
    CIoInteger          m_io_Data_UBA3_Out_iBurnerPowerMax              ;
    CIoInteger          m_io_Data_UBA3_Out_iBurnerPower                 ;
    CIoDouble           m_io_Data_UBA3_Out_dWaterTemp2                  ;
    CIoDouble           m_io_Data_UBA3_Out_dWaterTemp3                  ;
    CIoDouble           m_io_Data_UBA3_Out_dWorkingHours                ;
    CIoDouble           m_io_Data_UBA3_Out_dBurnerWorkingHours          ;
    CIoInteger          m_io_Data_UBA3_Out_iErrorCode                   ;
    CIoDouble           m_io_Data_UBA3_Out_dWaterTempNom                ;
    CIoDouble           m_io_Data_UBA3_Out_dWaterWorkingHours           ;
    CIoInteger          m_io_Data_UBA3_Out_iMode                        ;
    CIoBoolean          m_io_Data_UBA3_Out_bModeDay                     ;
    CIoBoolean          m_io_Data_UBA3_Out_bModeHeating                 ;
    CIoBoolean          m_io_Data_UBA3_Out_bModeNormalOperation         ;
    CIoInteger          m_io_Data_UBA3_Out_iPumpModulation              ;

    CIoDouble           m_io_Data_WM10_Out_dFlowTemperature1            ;
    CIoDouble           m_io_Data_WM10_Out_dFlowTemperature2            ;

    double              m_dLastUpdateTimePowerConsumption               ;
    double              m_dBurnerPowerOld                               ;
    bool                m_bBurnerActiveOld                              ;
    double              m_dNextIoSaveTime                               ;

    CIoDouble           m_io_Data_In_PowerConsumptionFactor_Water       ;
    CPersistentIo<CIoDouble, double> m_oPersistent_Data_In_PowerConsumptionFactor_Water;

    CIoDouble           m_io_Data_In_PowerConsumptionFactor_Heating     ;
    CPersistentIo<CIoDouble, double> m_oPersistent_Data_In_PowerConsumptionFactor_Heating;

    CIoDouble           m_io_Data_Out_PowerConsumption_All_Today        ;
    CPersistentIo<CIoDouble, double> m_oPersistent_Data_Out_PowerConsumption_All_Today;

    CIoDouble           m_io_Data_Out_PowerConsumption_All_Yesterday        ;
    CPersistentIo<CIoDouble, double> m_oPersistent_Data_Out_PowerConsumption_All_Yesterday;

    CIoDouble           m_io_Data_Out_PowerConsumption_All              ;
    CPersistentIo<CIoDouble, double> m_oPersistent_Data_Out_PowerConsumption_All;


    CIoDouble           m_io_Data_Out_PowerConsumption_Water_Today      ;
    CPersistentIo<CIoDouble, double> m_oPersistent_Data_Out_PowerConsumption_Water_Today;

    CIoDouble           m_io_Data_Out_PowerConsumption_Water_Yesterday      ;
    CPersistentIo<CIoDouble, double> m_oPersistent_Data_Out_PowerConsumption_Water_Yesterday;

    CIoDouble           m_io_Data_Out_PowerConsumption_Water_All        ;
    CPersistentIo<CIoDouble, double> m_oPersistent_Data_Out_PowerConsumption_Water_All;

    CIoDouble           m_io_Data_Out_PowerConsumption_Heating_Today    ;
    CPersistentIo<CIoDouble, double> m_oPersistent_Data_Out_PowerConsumption_Heating_Today;

    CIoDouble           m_io_Data_Out_PowerConsumption_Heating_Yesterday    ;
    CPersistentIo<CIoDouble, double> m_oPersistent_Data_Out_PowerConsumption_Heating_Yesterday;

    CIoDouble           m_io_Data_Out_PowerConsumption_Heating_All      ;
    CPersistentIo<CIoDouble, double> m_oPersistent_Data_Out_PowerConsumption_Heating_All;

    CIoStringPersist    m_io_Data_Out_PowerConsumption_strResetDate     ;
    CPersistentIo<CIoStringPersist, CString> m_oPersistent_Data_Out_PowerConsumption_strResetDate;
};

#endif /* _PLCEMSGATEWAY_H_ */
